import asyncio
import logging
from typing import Callable, Coroutine

from aiohttp import ClientSession

logger = logging.getLogger(__name__)


ATTEMPT_IP = [
    "1.0.0.1",  # Cloudflare
    "1.1.1.1",  # Cloudflare
    "208.67.222.222",  # OpenDNS
    "208.67.220.220",  # OpenDNS
]

is_connected = False
is_internet_checked = asyncio.Event()

connected_functions: list[Callable[[], Coroutine]] = []
disconnected_functions: list[Callable[[], Coroutine]] = []


def internet_connected():
    if is_internet_checked.is_set():
        return is_connected
    else:
        raise RuntimeError("Internet connection is not being monitored")


async def monitor_internet():
    global is_connected
    while True:
        # Try to connect to DNS servers and analyze internet connection
        was_connected = is_connected
        analyzed = False
        async with ClientSession() as session:
            for attempt_ip in ATTEMPT_IP:
                try:
                    async with session.get(f"http://{attempt_ip}") as response:
                        if response.status == 200:
                            analyzed = True
                            break
                except Exception:
                    pass
        is_connected = analyzed
        is_internet_checked.set()

        # Detect changes
        if was_connected and not is_connected:
            for job in disconnected_functions:
                asyncio.create_task(job())
            logger.warning("Internet disconnected")
        elif not was_connected and is_connected:
            for job in connected_functions:
                asyncio.create_task(job())
            logger.info("Internet connected")

        # Wait for a while
        await asyncio.sleep(1)


def when_internet_connected(job: Callable[[], Coroutine]):
    global connected_functions
    connected_functions.append(job)


def when_internet_disconnected(job: Callable[[], Coroutine]):
    global disconnected_functions
    disconnected_functions.append(job)



from collections import deque
from typing import TypeVar

F = TypeVar("F")
K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")


def list_to_dict(input_list: list[dict[F, K]], key_for_key: F) -> dict[K, dict[F, K]]:
    new_dict = {}
    for item in input_list:
        if key_for_key in item:
            key = item[key_for_key]
            new_dict[key] = item
    return new_dict


def slice_deque(original: deque[T], size: int, front: bool = False) -> list[T]:
    """
    Efficiently slices a `deque` from the specified size.
    """
    if front:
        # Slice from the beginning
        sliced: list[T] = []
        current_index = 0
        for each in original:
            if current_index >= size:
                break
            sliced.append(each)
            current_index += 1
    else:
        # Slice from the end
        sliced: list[T] = []
        current_index = 0
        for each in reversed(original):
            if current_index >= size:
                break
            sliced.append(each)
            current_index += 1
        sliced.reverse()

    return sliced



import asyncio
import json
import logging
from typing import Callable, Coroutine

from aiohttp import ClientError, ClientSession, WSMsgType

logger = logging.getLogger(__name__)


class ApiStreamError(Exception):
    def __init__(self, received: dict | list):
        formatted = json.dumps(received, indent=2)
        super().__init__(formatted)


class ApiStreamer:
    def __init__(
        self,
        url: str,
        handler: Callable[[dict], Coroutine] | Callable[[list], Coroutine],
    ):
        self._url = url
        self._handler = handler
        self._session = ClientSession()
        self._is_open = True

        asyncio.create_task(self._keep_connecting())

    @property
    def url(self) -> str:
        return self._url

    async def _keep_connecting(self):
        while self._is_open:
            try:
                await self._keep_listening()
            except ClientError:
                # This happens when internet is disconnected, etc...
                pass
            await asyncio.sleep(5.0)

    async def _keep_listening(self):
        async with self._session.ws_connect(self._url, heartbeat=5.0) as websocket:
            logger.info(f"Websocket connected\n{self._url}")
            async for message in websocket:
                if message.type == WSMsgType.ERROR:
                    url = self._url
                    parsed = json.dumps(message.json(), indent=2)
                    logger.warning(f"Websocket got an error message\n{url}\n{parsed}")
                else:
                    content = message.json()

                    def done_callback(task: asyncio.Task, content=content):
                        error = task.exception()
                        if error:
                            raise ApiStreamError(content) from error

                    task = asyncio.create_task(self._handler(content))
                    task.add_done_callback(done_callback)
            logger.info(f"Websocket disconnected\n{self._url}")

    async def close(self):
        self._is_open = False
        await self._session.close()
